import traceback
from typing import Any, Callable, Dict, List

from fastapi.logger import logger


class Action:
    def __init__(
        self,
        func: Callable,
        args: List[Any] = [],
        kwargs: Dict[str, Any] = {},
        name: str = "",
    ):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.name = name

    def do(self):
        self.func(*self.args, **self.kwargs)

    def __repr__(self):
        return (
            f"Action(func={self.func}, args={self.args}, "
            + f"kwargs={self.kwargs}, name={self.name})"
        )


class UndoManager:
    def __init__(self, logging: bool = False):
        self.logging = logging
        self.logging = True
        self.reset()
        self._undo_in_progress = False
        self._redo_in_progress = False
        self._pause = False
        self._undo_in_progress_name = ""
        self._redo_in_progress_name = ""

    def debug_log(self, msg: str, force: bool = False):
        if self.logging or force:
            logger.debug(msg)

    def pause(self):
        self._pause = True

    def unpause(self):
        self._pause = False

    def reset(self):
        "Clear both the undo and redo stacks"
        self.debug_log(f"Resetting undo/redo stacks")
        self.undo_actions: List[Action] = []
        self.redo_actions: List[Action] = []

    def add_action(
        self,
        func: Callable,
        args: List[Any] = [],
        kwargs: Dict[str, Any] = {},
        name: str = "",
    ):
        "Push a new action onto the undo stack."

        if self._pause:
            return

        action = Action(func, args=args, kwargs=kwargs, name=name)

        if self._undo_in_progress and self._redo_in_progress:
            self.debug_log(
                f"Undo action and Redo action in progress,"
                + f" ignoring new action: {name}"
            )
            return

        if self._undo_in_progress:
            action.name = self._undo_in_progress_name
            self.debug_log(
                f"Undo action in progress, adding new action to redo: {name}"
            )
            self.redo_actions.append(action)
            return

        if self._redo_in_progress:
            action.name = self._redo_in_progress_name
            self.debug_log(
                f"Redo action in progress, adding new action to undo: {name}"
            )
            self.undo_actions.append(action)
            return

        self.undo_actions.append(action)

    @property
    def can_undo(self) -> bool:
        "Is there an undo action in the stack"
        return len(self.undo_actions) > 0

    @property
    def can_redo(self) -> bool:
        "Is there an redo action in the stack"
        return len(self.redo_actions) > 0

    def undo(self):
        """
        Perform the next action in the undo stack and move it to
        the redo stack.
        """
        if self.can_undo:
            self.debug_log(f"Attempting undo for {self.next_undo_name}")
            action = self.undo_actions.pop()
            try:
                self._undo_in_progress = True
                self._undo_in_progress_name = action.name
                action.do()
                self._undo_in_progress = False
                self.debug_log(f"Undo complete")
            except Exception as e:  # noqa
                tb = traceback.format_exc()
                self.debug_log(f"Undo failed: {e}\n{tb}", force=True)
                self.reset()
                return
        else:
            self.debug_log(f"Unable to undo. Undo stack is empty!")

    def redo(self):
        """
        Perform the next action in the redo stack and move it to
        the undo stack.
        """
        if self.can_redo:
            self.debug_log(f"Attempting redo for {self.next_redo_name}")
            action = self.redo_actions.pop()
            try:
                self._redo_in_progress = True
                self._redo_in_progress_name = action.name
                action.do()
                self._redo_in_progress = False
                self.debug_log(f"Redo complete")
            except Exception as e:  # noqa
                tb = traceback.format_exc()
                self.debug_log(f"Redo failed: {e}\n{tb}", force=True)
                self.reset()
                return
        else:
            self.debug_log(f"Unable to redo. Redo stack is empty!")

    @property
    def undo_count(self) -> int:
        "Number of actions in the undo stack."
        return len(self.undo_actions)

    @property
    def redo_count(self) -> int:
        "Number of actions in the redo stack."
        return len(self.redo_actions)

    @property
    def next_undo_name(self) -> str:
        "Name of the next action in the undo stack"
        if self.can_undo:
            return self.undo_actions[-1].name
        else:
            return ""

    @property
    def next_redo_name(self) -> str:
        "Name of the next action in the redo stack"
        if self.can_redo:
            return self.redo_actions[-1].name
        else:
            return ""



import asyncio
import logging
import sys

from asyncblink import signal  # type: ignore
from fastapi import Depends, FastAPI, Request
from fastapi.logger import logger
from fastapi.middleware.wsgi import WSGIMiddleware
from starlette.responses import JSONResponse
from uvicorn.main import Server  # type: ignore

from jupyter_d1.signals import APP_LOG, APP_SHUTDOWN, APP_STARTUP

from .d1_response import D1Response
from .dav import dav, dav_manager
from .deps import read_access
from .routers import convert, hello, kernels, login, notebooks, rclone, server
from .settings import settings
from .storage import stats_manager
from .storage.notebook_manager import NBMException

# Monkey patch the uvicorn exit handler. The standard
# uvicorn 'shutdown' event comes too late to properly
# shutdown active websockets. There are no other hooks
# so this is the only way to do so.
# See https://stackoverflow.com/a/59887076/436040
uvicorn_handle_exit = Server.handle_exit


def handle_exit(self, *args, **kwargs):
    logger.debug("Sending APP_SHUTDOWN signal")
    signal(APP_SHUTDOWN).send()
    uvicorn_handle_exit(self, *args, **kwargs)


Server.handle_exit = handle_exit


class AsyncBlinkHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level=level)

    def emit(self, record):
        msg = self.format(record)
        signal(APP_LOG).send(msg=msg)


formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - " "%(message)s"
)
handler = AsyncBlinkHandler()
handler.setFormatter(formatter)
logger.setLevel(logging.DEBUG)
logger.addHandler(handler)

logging.getLogger("wsgidav").addHandler(logging.StreamHandler(sys.stdout))

app = FastAPI(default_response_class=D1Response)
app.mount("/dav", WSGIMiddleware(dav))

# Begin FastAPI setup

rclone_stats_task = None
server_stats_task = None


@app.on_event("startup")
def startup_event():
    logger.debug(f"App startup")
    signal(APP_STARTUP).send()
    loop = asyncio.get_event_loop()
    if settings.RCLONE_STATS_POLL:
        global rclone_stats_task
        global server_stats_task
        rclone_stats_task = loop.create_task(rclone.stats_periodic())
        server_stats_task = loop.create_task(
            stats_manager.server_stats_periodic()
        )
    if settings.WATCHDOG_ENABLED:
        dav_manager.start_watchdog(loop)


@app.on_event("shutdown")
async def shutdown_event():
    logger.debug("App shutdown")
    try:
        rclone_stats_task.cancel()
    except Exception:
        pass

    try:
        server_stats_task.cancel()
    except Exception:
        pass

    dav_manager.stop_watchdog()


not_found_resp = {
    "error": "NOT_FOUND",
    "reason": "Not found.",
    "detail": None,
}

app.include_router(
    hello.router,
    prefix="/hello",
    tags=["hello"],
    responses={404: not_found_resp},
)

app.include_router(
    notebooks.router,
    prefix="/notebooks",
    tags=["notebooks"],
    dependencies=[Depends(read_access)],
    responses={404: not_found_resp},
)

app.include_router(
    convert.router,
    prefix="/convert",
    tags=["convert"],
    dependencies=[Depends(read_access)],
    responses={404: not_found_resp},
)

app.include_router(
    kernels.router,
    prefix="/kernels",
    tags=["kernels"],
    dependencies=[Depends(read_access)],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    rclone.router,
    prefix="/rclone",
    tags=["rclone"],
    dependencies=[Depends(read_access)],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    server.router,
    prefix="/server",
    tags=["server"],
    dependencies=[Depends(read_access)],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    login.router,
    tags=["login"],
    responses={404: {"description": "Not found"}},
)


@app.exception_handler(NBMException)
async def notebook_manager_exception_handler(
    request: Request, exc: NBMException
):

    return JSONResponse(
        status_code=404,
        content={
            "error": exc.error_code,
            "reason": exc.message,
            "detail": None,
        },
    )



import asyncio
import os
from typing import Any, Dict, Tuple

from fastapi.logger import logger
from jupyter_client import AsyncMultiKernelManager  # type: ignore

from jupyter_core.paths import jupyter_runtime_dir  # type: ignore

from .kernel_listener import KernelListener

try:
    from nb_conda_kernels import CondaKernelSpecManager as KernelSpecManager
    logger.debug("Successfully loaded nb_conda_kernels.")
except ImportError as e:
    from jupyter_client.kernelspec import KernelSpecManager
    logger.debug(f"nb_conda_kernels disabled.  Failed to import - {e}")


class KernelManager:
    def __init__(self):
        self._specManager = KernelSpecManager()
        self._kmanager = AsyncMultiKernelManager(kernel_spec_manager=self._specManager)
        self._kmanager.connection_dir = jupyter_runtime_dir()
        if not os.path.exists(self._kmanager.connection_dir):
            os.makedirs(
                self._kmanager.connection_dir, mode=0o700, exist_ok=True
            )

        self.clients = {}  # map of notebook uuid to client objects
        self.listeners = {}

    async def shutdown_all(self):
        # make a copy so we don't change the array during iteration
        uuids = list(self.clients.keys())

        for uuid in uuids:
            await self.shutdown_kernel(uuid)

    async def shutdown_kernel(self, uuid):
        listener = self.listeners.get(uuid)
        if listener is not None:
            await listener.shutdown_listener()
        del self.clients[uuid]
        await self._kmanager.shutdown_kernel(str(uuid))

    async def restart_kernel(self, uuid):
        await self._kmanager.restart_kernel(str(uuid))

    async def interrupt_kernel(self, uuid):
        await self._kmanager.interrupt_kernel(str(uuid))

    async def kernel_is_alive(self, uuid) -> bool:
        return await self._kmanager.is_alive(str(uuid))

    async def start_kernel(
        self, kernel_name, uuid, directory, kernel_options=None
    ):
        if uuid in self.clients.keys():
            logger.info(f"replacing kernel for {uuid}")
            await self.shutdown_kernel(uuid)
        else:
            logger.info(f"starting kernel for {uuid}")
        if kernel_options is None:
            kernel_options = []
        await self._kmanager.start_kernel(
            kernel_name=kernel_name,
            kernel_id=str(uuid),
            cwd=directory,
            extra_arguments=kernel_options,
        )
        kernel = self._kmanager.get_kernel(str(uuid))

        client = kernel.client()
        client.start_channels()
        self.clients[uuid] = client

        # start the event listener
        listener = KernelListener(client, uuid)
        self.listeners[uuid] = listener
        asyncio.create_task(listener.listen())

    async def execute(
        self,
        uuid: str,
        code: str,
        silent: bool = False,
        store_history: bool = True,
    ):
        client = self.clients[uuid]
        msg_id = client.execute(
            code, silent=silent, store_history=store_history,
        )
        return msg_id

    async def complete(self, uuid: str, code: str, cursor_pos: int = None):
        client = self.clients[uuid]
        msg_id = client.complete(code, cursor_pos=cursor_pos)
        return msg_id

    async def get_history(self, uuid: str):
        client = self.clients[uuid]
        msg_id = client.history(output=True)
        return msg_id

    def get_all_kernelspecs(self):
        return self._specManager.get_all_specs()

    def kernel_names(self):
        kspecs = self._specManager.get_all_specs()
        return list(kspecs.keys())

    def get_kernelspec(self, uuid):
        return self._kmanager.get_kernel(str(uuid)).kernel_spec

    def get_kernelspec_by_name(self, name) -> Tuple[str, Dict[str, Any]]:
        all_specs = self._specManager.get_all_specs()
        if name not in all_specs:
            # if there's no matching kernel_name, default to a `python` kernel
            name = list(
                filter(lambda x: "python" in x, list(all_specs.keys()))
            ).pop()

        return name, all_specs[name]



import imp
import json
import os
import platform
import subprocess
import sys

from fastapi.logger import logger

try:
    import psutil  # type: ignore

    has_psutil = True
    logger.debug("Successfully loaded psutil.")
except ImportError:
    has_psutil = False
    logger.debug("Failed to load psutil.  Disabling.")
from typing import Any, Dict, Optional

from ..models.environment_info import (
    CondaInfo,
    EnvironmentInfo,
    OSInfo,
    ProcessInfo,
    PythonInfo,
    PythonVersionInfo,
)
from ..settings import settings

# Based on
# https://github.com/Runscope/healthcheck/blob/master/healthcheck/__init__.py


class EnvironmentDump(object):
    def dump_environment(self) -> EnvironmentInfo:
        if has_psutil:
            cpu_count = psutil.cpu_count()
        else:
            cpu_count = 0
        return EnvironmentInfo(
            os=self.get_os(),
            python=self.get_python(),
            config=self.get_config(),
            process=self.get_process(),
            conda=self.get_conda(),
            cores=cpu_count,
            version=settings.VERSION,
        )

    def get_os(self) -> OSInfo:
        raw_name = platform.system()
        name = raw_name
        if raw_name.startswith("Darwin"):
            name = "macOS"
        return OSInfo(
            platform=sys.platform,
            name=name,
            raw_name=raw_name,
            uname=dict(platform.uname()._asdict()),
        )

    def get_config(self) -> Dict[str, Any]:
        return self.safe_dump(settings.dict())

    def get_python(self) -> PythonInfo:
        python_info = PythonInfo(
            version=sys.version,
            executable=sys.executable,
            pythonpath=sys.path,
            version_info=PythonVersionInfo(
                major=sys.version_info.major,
                minor=sys.version_info.minor,
                micro=sys.version_info.micro,
                releaselevel=sys.version_info.releaselevel,
                serial=sys.version_info.serial,
            ),
        )
        if imp.find_module("pkg_resources"):
            import pkg_resources

            packages = dict(
                [
                    (p.project_name, p.version)
                    for p in pkg_resources.working_set
                ]
            )
            python_info.packages = packages

        return python_info

    def get_conda(self) -> Optional[CondaInfo]:
        try:
            process = subprocess.run(
                ["conda", "--version"], stdout=subprocess.PIPE
            )
            version = process.stdout.decode("utf-8")
            return CondaInfo(version=version)
        except Exception:
            return None

    def get_login(self):
        # Based on https://github.com/gitpython-developers/GitPython/pull/43/
        # Fix for 'Inappopropirate ioctl for device' on posix systems.
        if os.name == "posix":
            import pwd

            username = pwd.getpwuid(os.geteuid()).pw_name
        else:
            username = os.environ.get(
                "USER", os.environ.get("USERNAME", "UNKNOWN")
            )
            if username == "UNKNOWN" and hasattr(os, "getlogin"):
                username = os.getlogin()
        return username

    def get_process(self) -> ProcessInfo:
        return ProcessInfo(
            argv=sys.argv,
            cwd=os.getcwd(),
            user=self.get_login(),
            pid=os.getpid(),
            environ=self.safe_dump(os.environ),
        )

    def safe_dump(self, dictionary) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        for key in dictionary.keys():
            if (
                key.lower().endswith("key")
                or key.lower().endswith("token")
                or key.lower().endswith("pass")
            ):
                # Try to avoid listing passwords and access tokens or
                # keys in the output
                result[key] = "********"
            else:
                try:
                    json.dumps(dictionary[key])
                    result[key] = dictionary[key]
                except TypeError:
                    pass
        return result



from typing import List, Optional

from pydantic import BaseModel

from .base_wrapper import BaseWrapper
from .JSONType import JSONType


class KernelVariable(BaseModel):
    name: str
    type: Optional[str]
    abbreviated: Optional[bool] = False
    value: JSONType
    summary: Optional[str]


class KernelVariableWrapper(BaseWrapper):
    single_var: KernelVariable


class KernelVariablesWrapper(BaseWrapper):
    vars: List[KernelVariable]



from uvicorn.config import LOGGING_CONFIG
from uvicorn.workers import UvicornWorker

LOGGING_CONFIG["formatters"]["default"][
    "fmt"
] = '%(asctime)s %(levelprefix)s - %(message)s'
LOGGING_CONFIG["formatters"]["access"][
    "fmt"
] = '%(asctime)s %(levelprefix)s - "%(request_line)s" %(status_code)s'


class D1UvicornWorker(UvicornWorker):
    CONFIG_KWARGS = {"loop": "auto", "http": "auto", "log_config": LOGGING_CONFIG}


from uuid import UUID

from fastapi import APIRouter, Depends, status

from ..deps import write_access
from ..models.kernel import KernelIsAlive, KernelsIdle
from ..models.kernel_spec import KernelSpec, KernelSpecsWrapper
from ..storage import kmanager, manager

router = APIRouter()


@router.get("/specs", response_model=KernelSpecsWrapper)
async def list_kernel_specs():
    "Return a list of all kernel specs available on the server."
    kspecs = kmanager.get_all_kernelspecs()
    specs = []
    for key in kspecs.keys():
        tmp_spec = kspecs[key]
        tmp_spec["kernel_name"] = key
        new_spec = KernelSpec(**tmp_spec)
        specs.append(new_spec)
    return KernelSpecsWrapper(kernel_specs=specs)


@router.get(
    "/shutdown/{uuid}",
    dependencies=[Depends(write_access)],
    status_code=status.HTTP_204_NO_CONTENT,
)
async def shutdown_kernel(uuid: UUID):
    "Shutdown the kernel identified by the uuid"
    await kmanager.shutdown_kernel(uuid)


@router.get(
    "/restart/{uuid}",
    dependencies=[Depends(write_access)],
    status_code=status.HTTP_204_NO_CONTENT,
)
async def restart_kernel(uuid: UUID):
    "Restart the kernel identified by the uuid"
    await kmanager.restart_kernel(uuid)


@router.get(
    "/interrupt/{uuid}",
    dependencies=[Depends(write_access)],
    status_code=status.HTTP_204_NO_CONTENT,
)
async def interrupt_kernel(uuid: UUID):
    "Interrupt the kernel identified by the uuid"
    await kmanager.interrupt_kernel(uuid)


@router.get("/is_alive/{uuid}", response_model=KernelIsAlive)
async def kernel_is_alive(uuid: str, status_code=status.HTTP_204_NO_CONTENT):
    "Check that kernel process is running."
    is_alive = kmanager.kernel_is_alive(uuid)
    return KernelIsAlive(kernel_is_alive=is_alive, kernel_id=uuid)


@router.get(
    "/idle", status_code=status.HTTP_200_OK, response_model=KernelsIdle
)
def idle_status():
    "Check if all kernels are idle"
    return KernelsIdle(idle=manager.is_idle(), last_idle=manager.last_idle)



from enum import Enum
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel

from .JSONType import JSONType
from .kernel_message import KernelMessageWrapper


class CellType(str, Enum):
    code = "code"
    markdown = "markdown"
    raw = "raw"


class CellWrapper(BaseModel):
    cell: JSONType


class CellKernelMessageWrapper(CellWrapper, KernelMessageWrapper):
    pass


class CellsWrapper(BaseModel):
    cells: List[JSONType]


class CellUpdate(BaseModel):
    source: Optional[str] = None
    cell_type: Optional[CellType] = None


class CellAdd(BaseModel):
    source: str = ""
    cell_type: CellType = CellType.code
    before: Optional[UUID] = None



from datetime import datetime, timedelta
from typing import Any, Union

from jose import jwt  # type: ignore

from .settings import settings

ALGORITHM = "HS256"


def create_access_token(
    subject: Union[str, Any], expires_delta: timedelta = None
) -> str:
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(
            minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES
        )
    to_encode = {"exp": expire, "sub": str(subject)}
    encoded_jwt = jwt.encode(
        to_encode, settings.SECRET_KEY, algorithm=ALGORITHM
    )
    return encoded_jwt



import asyncio
import os
import pathlib
import platform
import shutil
from typing import Dict
from unittest import TestCase

import pytest  # type: ignore
from async_asgi_testclient import (
    TestClient as WebSocketTestClient,  # type: ignore
)
from fastapi import WebSocket
from fastapi.testclient import TestClient
from pytest_mock import MockFixture  # type: ignore

from jupyter_d1.settings import settings

from .test_notebooks_websocket import upload_notebook
from .utils import (
    get_permissionless_token,
    get_read_only_token,
    get_superuser_token,
    wait_for_event,
    workdir_1,
)

is_mac = platform.system().startswith("Darwin")


class TestDav:
    def test_dav_auth(
        self, client: TestClient,
    ):
        response = client.get("/dav")
        assert response.status_code == 401

        response = client.get(f"/dav/", auth=(get_permissionless_token(), ""))
        assert response.status_code == 401

        # Alias of / to ROOT_DIR was removed, make sure it doesn't work
        response = client.get(f"/dav/", auth=(get_read_only_token(), ""))
        assert response.status_code == 404

        response = client.get(
            f"/dav{settings.ROOT_DIR}", auth=(get_read_only_token(), "")
        )
        assert response.status_code == 200

        response = client.get(
            f"/dav{settings.ROOT_DIR}", auth=(get_superuser_token(), "")
        )
        assert response.status_code == 200

    def test_dav_provider_added_for_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        shutil.copyfile(
            "jupyter_d1/tests/notebooks/simple.ipynb",
            "jupyter_d1/tests/notebooks/simple_copy.ipynb",
        )
        other_directory = f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        token = get_read_only_token()
        response = client.get(f"/dav{other_directory}", auth=(token, ""))
        assert response.status_code == 404

        path = f"{other_directory}/simple_copy.ipynb"
        response = client.get(
            f"/notebooks/open/?filepath={path}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        assert (
            resp_json["metadata"]["jupyter_d1"]["working_directory"]
            == other_directory
        )
        response = client.get(f"/dav{other_directory}", auth=(token, ""))
        assert response.status_code == 200

        response = client.delete(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(f"/dav{other_directory}", auth=(token, ""))
        assert response.status_code == 404

        os.remove("jupyter_d1/tests/notebooks/simple_copy.ipynb")

    def test_dav_provider_added_for_notebook_set_working_dir(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        shutil.copyfile(
            "jupyter_d1/tests/notebooks/simple.ipynb",
            "jupyter_d1/tests/notebooks/simple_copy.ipynb",
        )
        other_directory = f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        working_dir = "/tmp"
        resolved_work_dir = str(pathlib.Path(working_dir).resolve())
        token = get_read_only_token()
        response = client.get(f"/dav{resolved_work_dir}", auth=(token, ""))
        assert response.status_code == 404

        path = f"{other_directory}/simple_copy.ipynb"
        response = client.get(
            f"/notebooks/open/?filepath={path}",
            headers=superuser_token_headers,
            params={"working_directory": working_dir},
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        assert (
            resp_json["metadata"]["jupyter_d1"]["working_directory"]
            == resolved_work_dir
        )
        response = client.get(f"/dav{resolved_work_dir}", auth=(token, ""))
        assert response.status_code == 200

        response = client.delete(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(f"/dav{resolved_work_dir}", auth=(token, ""))
        assert response.status_code == 404

        os.remove("jupyter_d1/tests/notebooks/simple_copy.ipynb")

    def test_dav_provider_notebook_in_subdirectory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        If notebook is in subdirectory of d1's working directory, 
        dav should work regardless of whether the notebook is open.
        """
        pathlib.Path(f"{settings.ROOT_DIR}/some_nbs").mkdir(exist_ok=True)
        shutil.copyfile(
            "jupyter_d1/tests/notebooks/simple.ipynb",
            f"{settings.ROOT_DIR}/some_nbs/simple_copy.ipynb",
        )

        nb_dir = f"{settings.ROOT_DIR}/some_nbs"
        working_dir = f"{settings.ROOT_DIR}/some_nbs"
        token = get_read_only_token()
        response = client.get(f"/dav{working_dir}", auth=(token, ""))
        assert response.status_code == 200

        path = f"{nb_dir}/simple_copy.ipynb"
        response = client.get(
            f"/notebooks/open/?filepath={path}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        assert (
            resp_json["metadata"]["jupyter_d1"]["working_directory"]
            == working_dir
        )
        response = client.get(f"/dav{working_dir}", auth=(token, ""))
        assert response.status_code == 200

        response = client.delete(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(f"/dav{working_dir}", auth=(token, ""))
        assert response.status_code == 200

        shutil.rmtree(f"{settings.ROOT_DIR}/some_nbs")


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "clear_notebook_directory", "clear_notebook_directory_2"
)
class TestWatchdogWebSocket:
    async def clear_websocket_queue(self, websocket, wait_for=0.3):
        while True:
            try:
                await asyncio.wait_for(websocket.receive_json(), wait_for)
            except Exception:
                break

    async def create_dir(self, websocket: WebSocket, ututils: TestCase):
        os.mkdir(f"{settings.ROOT_DIR}/new_dir")

        resp = await wait_for_event(websocket, "watchdog_event")
        resp2 = await wait_for_event(websocket, "watchdog_event")

        ututils.assertCountEqual(
            [resp, resp2],
            [
                {
                    "watchdog_event": {
                        "event_type": "created",
                        "src_path": f"{settings.ROOT_DIR}/new_dir",
                        "dest_path": None,
                    }
                },
                {
                    "watchdog_event": {
                        "event_type": "modified",
                        "src_path": settings.ROOT_DIR,
                        "dest_path": None,
                    }
                },
            ],
        )

    async def create_file(
        self,
        websocket: WebSocket,
        ututils: TestCase,
        directory: str = settings.ROOT_DIR,
    ):
        with open(f"{directory}/new_file", "w") as f:
            f.write("this is a new file")

        resp1 = await wait_for_event(websocket, "watchdog_event")
        resp2 = await wait_for_event(websocket, "watchdog_event")
        expected = [
            {
                "watchdog_event": {
                    "event_type": "modified",
                    "src_path": f"{directory}",
                    "dest_path": None,
                }
            },
            {
                "watchdog_event": {
                    "event_type": "created",
                    "src_path": f"{directory}/new_file",
                    "dest_path": None,
                }
            },
        ]
        ututils.assertCountEqual([resp1, resp2], expected)

    async def test_create_dir(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
        ututils: TestCase,
        mocker: MockFixture,
    ):
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            # assert False
            await self.create_dir(websocket, ututils)

    async def test_create_file(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
        ututils: TestCase,
    ):
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            await self.create_file(websocket, ututils)

    async def test_modified_file(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
        ututils: TestCase,
    ):
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            await self.clear_websocket_queue(websocket)
            await self.create_file(websocket, ututils)

            with open(f"{settings.ROOT_DIR}/new_file", "w") as f:
                f.write("this is an edit")
            expected = [
                {
                    "watchdog_event": {
                        "event_type": "modified",
                        "src_path": f"{settings.ROOT_DIR}/new_file",
                        "dest_path": None,
                    }
                }
            ]
            resps = []
            for _ in expected:
                resps.append(await wait_for_event(websocket, "watchdog_event"))
            for resp in resps:
                assert expected.pop(expected.index(resp)) is not None

    async def test_deleted_file(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
        ututils: TestCase,
    ):
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            await self.clear_websocket_queue(websocket)
            await self.create_file(websocket, ututils)
            await self.clear_websocket_queue(websocket)

            os.remove(f"{settings.ROOT_DIR}/new_file")

            expected = [
                {
                    "watchdog_event": {
                        "event_type": "modified",
                        "src_path": settings.ROOT_DIR,
                        "dest_path": None,
                    }
                },
                {
                    "watchdog_event": {
                        "event_type": "deleted",
                        "src_path": f"{settings.ROOT_DIR}/new_file",
                        "dest_path": None,
                    }
                },
            ]
            resps = []
            resp1 = await wait_for_event(websocket, "watchdog_event")
            resp2 = await wait_for_event(websocket, "watchdog_event")
            resps = [resp1, resp2]
            # This modified event doesn't appear to happen on linux
            if is_mac:
                resp3 = await wait_for_event(websocket, "watchdog_event")
                resps.append(resp3)
                expected.append(
                    {
                        "watchdog_event": {
                            "event_type": "modified",
                            "src_path": f"{settings.ROOT_DIR}/new_file",
                            "dest_path": None,
                        }
                    }
                )
            ututils.assertCountEqual(resps, expected)

    async def test_nested_updates(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
        ututils: TestCase,
    ):
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            await self.clear_websocket_queue(websocket)
            await self.create_dir(websocket, ututils)
            await self.create_file(
                websocket, ututils, directory=f"{settings.ROOT_DIR}/new_dir"
            )
            await self.clear_websocket_queue(websocket)
            os.rename(
                f"{settings.ROOT_DIR}/new_dir/new_file",
                f"{settings.ROOT_DIR}/new_file",
            )

            expected = [
                {
                    "watchdog_event": {
                        "event_type": "moved",
                        "src_path": f"{settings.ROOT_DIR}/new_dir/new_file",
                        "dest_path": f"{settings.ROOT_DIR}/new_file",
                    }
                },
                {
                    "watchdog_event": {
                        "event_type": "modified",
                        "src_path": f"{settings.ROOT_DIR}/new_dir",
                        "dest_path": None,
                    }
                },
            ]

            # More differences between mac and linux
            if is_mac:
                expected.append(
                    {
                        "watchdog_event": {
                            "event_type": "modified",
                            "src_path": f"{settings.ROOT_DIR}/"
                            "new_dir/new_file",
                            "dest_path": None,
                        }
                    }
                )
            else:
                expected.append(
                    {
                        "watchdog_event": {
                            "event_type": "modified",
                            "src_path": "/tmp/jupyter_d1_test",
                            "dest_path": None,
                        }
                    }
                )
            resps = []
            for _ in expected:
                resps.append(await wait_for_event(websocket, "watchdog_event"))
            ututils.assertCountEqual(resps, expected)

    async def test_added_for_notebook(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
    ):
        """
        Monitor new workdir for notebook, and stop monitoring when changed.
        """
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            new_workdir = workdir_1

            # Should not get notified of directory changes yet
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir")
            with pytest.raises(AssertionError):
                await wait_for_event(websocket, "watchdog_event")

            # Upload notebook, workdir same as root dir
            uuid = await upload_notebook(
                websocket_client, superuser_token_headers,
            )

            # Change nb workdir
            response = await websocket_client.patch(
                f"/notebooks/{uuid}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": new_workdir},
            )
            assert response.status_code == 200

            # Make sure new nb workdir is now monitored for changes
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir2")
            event = await wait_for_event(websocket, "watchdog_event")
            assert event == {
                "watchdog_event": {
                    "event_type": "created",
                    "src_path": f"{new_workdir}/new_dir2",
                    "dest_path": None,
                }
            }

            # Change nb workdir back to root dir
            response = await websocket_client.patch(
                f"/notebooks/{uuid}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": settings.ROOT_DIR},
            )
            assert response.status_code == 200

            # Make sure previous nb workdir is no longer monitored
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir3")
            with pytest.raises(AssertionError):
                await wait_for_event(websocket, "watchdog_event")

    async def test_change_monitored_directory(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
    ):
        """Monitor more specific directory when possible"""
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            new_workdir = workdir_1

            # Should not get notified of directory changes yet
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir")
            with pytest.raises(AssertionError):
                await wait_for_event(websocket, "watchdog_event")

            # Upload notebook, workdir same as root dir
            uuid = await upload_notebook(
                websocket_client, superuser_token_headers
            )

            # Change nb workdir
            response = await websocket_client.patch(
                f"/notebooks/{uuid}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": new_workdir},
            )
            assert response.status_code == 200

            # Make sure new nb workdir is now monitored for changes
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir2")
            event = await wait_for_event(websocket, "watchdog_event")
            assert event == {
                "watchdog_event": {
                    "event_type": "created",
                    "src_path": f"{new_workdir}/new_dir2",
                    "dest_path": None,
                }
            }

            # Change nb workdir to sub directory
            response = await websocket_client.patch(
                f"/notebooks/{uuid}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": f"{new_workdir}/new_dir"},
            )
            assert response.status_code == 200

            # Make sure parent dir is no longer monitored
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir3")
            with pytest.raises(AssertionError):
                await wait_for_event(websocket, "watchdog_event")

            # Make sure sub dir is now monitored for changes
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir/newer_dir")
            event = await wait_for_event(websocket, "watchdog_event")
            assert event == {
                "watchdog_event": {
                    "event_type": "created",
                    "src_path": f"{new_workdir}/new_dir/newer_dir",
                    "dest_path": None,
                }
            }

    async def test_multiple_notebooks(
        self,
        websocket_client: WebSocketTestClient,
        superuser_token_headers: Dict[str, str],
    ):
        """
        Multiple notebooks, one with workdir that is parent dir of the other's
        workdir. Then switch the parent workdir to a sub dir, make sure both
        subdirs are still monitored, but not the parent
        """
        async with websocket_client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:
            new_workdir = workdir_1

            # Should not get notified of directory changes yet
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir")
            with pytest.raises(AssertionError):
                await wait_for_event(websocket, "watchdog_event")

            # Upload notebook, workdir same as root dir
            uuid = await upload_notebook(
                websocket_client, superuser_token_headers,
            )
            uuid2 = await upload_notebook(
                websocket_client,
                superuser_token_headers,
                filename="other_simple.ipynb",
            )
            await asyncio.sleep(3.0)

            # Change nb workdir
            response = await websocket_client.patch(
                f"/notebooks/{uuid}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": new_workdir},
            )
            assert response.status_code == 200
            response = await websocket_client.patch(
                f"/notebooks/{uuid2}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": f"{new_workdir}/new_dir"},
            )
            assert response.status_code == 200

            # Make sure new nb workdir is now monitored for changes
            await self.clear_websocket_queue(websocket)
            await asyncio.sleep(10.0)
            os.mkdir(f"{new_workdir}/new_dir2")
            event = await wait_for_event(websocket, "watchdog_event")
            assert event == {
                "watchdog_event": {
                    "event_type": "created",
                    "src_path": f"{new_workdir}/new_dir2",
                    "dest_path": None,
                }
            }

            # Change nb workdir to sub directory
            response = await websocket_client.patch(
                f"/notebooks/{uuid}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": f"{new_workdir}/new_dir2"},
            )
            assert response.status_code == 200

            # Make sure parent dir is no longer monitored
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir3")
            with pytest.raises(AssertionError):
                await wait_for_event(websocket, "watchdog_event")

            # Make sure sub dir is still monitored for changes
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir/newer_dir")
            event = await wait_for_event(websocket, "watchdog_event")
            assert event == {
                "watchdog_event": {
                    "event_type": "created",
                    "src_path": f"{new_workdir}/new_dir/newer_dir",
                    "dest_path": None,
                }
            }

            # Make sure other sub dir is now also monitored for changes
            await self.clear_websocket_queue(websocket)
            os.mkdir(f"{new_workdir}/new_dir2/newer_dir2")
            event = await wait_for_event(websocket, "watchdog_event")
            assert event == {
                "watchdog_event": {
                    "event_type": "created",
                    "src_path": f"{new_workdir}/new_dir2/newer_dir2",
                    "dest_path": None,
                }
            }



import json
from os import path
from pathlib import Path
from typing import Any, Dict, List, Optional
from uuid import UUID

from asyncblink import signal  # type: ignore
from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    HTTPException,
    Request,
    Response,
    WebSocket,
    status,
)
from fastapi.encoders import jsonable_encoder
from fastapi.logger import logger
import jsonschema

from jupyter_d1.signals import (
    CELL_ADDED,
    CELL_DELETED,
    CELL_EXECUTION_INPUT,
    CELL_EXECUTION_REPLY,
    CELL_EXECUTION_REQUEST,
    CELL_STREAM,
    CELL_UPDATE,
    COMPLETE_REPLY,
    CONTROL_CHANNEL,
    HB_CHANNEL,
    HISTORY_REPLY,
    IOPUB_CHANNEL,
    KERNEL_INTERRUPTED,
    KERNEL_RESTARTED,
    METADATA_UPDATE,
    PAYLOAD_PAGE,
    SCRATCH_UPDATE,
    SHELL_CHANNEL,
    STDIN_CHANNEL,
    VARS_UPDATE,
)

from ..d1_response import D1Encoder, D1Response
from ..deps import read_access_websocket, write_access, write_access_websocket
from ..models.cell import (
    CellAdd,
    CellKernelMessageWrapper,
    CellsWrapper,
    CellUpdate,
    CellWrapper,
)
from ..models.code_complete import CodeComplete
from ..models.kernel_message import (
    KernelMessage,
    KernelMessagesWrapper,
    KernelMessageWrapper,
)
from ..models.kernel_variable import (
    KernelVariable,
    KernelVariablesWrapper,
    KernelVariableWrapper,
)
from ..models.notebook import NotebookPath, NotebooksWrapper, NotebookWrapper
from ..models.scratch import ScratchCode
from ..settings import settings
from ..storage import manager
from ..utils import NotebookNode, add_to_dict_metadata, websocket_poll

router = APIRouter(default_response_class=D1Response)


@router.get("", response_model=NotebooksWrapper)
async def get_notebooks():
    notebooks = manager.get_notebooks_dicts()
    return NotebooksWrapper(notebooks=notebooks)


@router.delete(
    "",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def delete_notebooks():
    await manager.delete_notebooks()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get("/{uuid}", response_model=NotebookWrapper)
async def get_notebook(uuid: UUID):
    notebook = manager.get_notebook_dict(uuid)
    return NotebookWrapper(notebook=notebook)


@router.delete(
    "/{uuid}",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def delete_notebook(uuid: UUID):
    await manager.delete_notebook(uuid)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get("/{uuid}/cells", response_model=CellsWrapper)
async def get_notebook_cells(uuid: UUID):
    node = manager.get_notebook_node(uuid)
    if node is not None:
        cells = node.cells
    else:
        raise HTTPException(
            status_code=404, detail=f"Notebook not found for id {uuid}"
        )
    return CellsWrapper(cells=cells)


def schedule_notebook_save(background_tasks: BackgroundTasks, uuid: UUID):
    if manager.get_notebook(uuid).autosave:
        background_tasks.add_task(manager.save_notebook, uuid=uuid)


@router.post(
    "/{uuid}/cells",
    status_code=status.HTTP_201_CREATED,
    response_model=CellWrapper,
    dependencies=[Depends(write_access)],
)
async def create_notebook_cell(
    uuid: UUID, cell_add: Optional[CellAdd], background_tasks: BackgroundTasks
) -> Any:
    """Create a new cell with type `cell_type` and source code of `source`.

    The `before` parameter is the uuid of the cell will be inserted
    before.
    If `before`
    is omitted, the cell will be added as the last cell.
    """
    if cell_add is None:
        cell_add = CellAdd()
    new_cell = manager.create_cell(
        notebook_id=uuid,
        cell_type=cell_add.cell_type,
        source=cell_add.source,
        before=cell_add.before,
    )
    schedule_notebook_save(background_tasks, uuid=uuid)
    return CellWrapper(cell=new_cell)


@router.get(
    "/{uuid}/cells/{cell_uuid}/move",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def move_notebook_cell(
    uuid: UUID,
    cell_uuid: UUID,
    background_tasks: BackgroundTasks,
    before: Optional[UUID] = None,
):
    """Move an existing cell with uuid `cell_uuid` to the slot before
    the cell with uuid `before`.
    If `before`
    is omitted, the cell will be moved to the end of the cell list.
    """
    manager.move_cell(notebook_id=uuid, cell_id=cell_uuid, before=before)
    schedule_notebook_save(background_tasks, uuid=uuid)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{uuid}/cells/{cell_uuid}/merge",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def merge_notebook_cells(
    uuid: UUID,
    cell_uuid: UUID,
    background_tasks: BackgroundTasks,
    above: bool = False,
):
    """Merge a cell with the cell above or below it."""
    manager.merge_cells(notebook_id=uuid, cell_id=cell_uuid, above=above)
    schedule_notebook_save(background_tasks, uuid=uuid)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{uuid}/cells/{cell_uuid}/split",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def split_notebook_cell(
    uuid: UUID,
    cell_uuid: UUID,
    background_tasks: BackgroundTasks,
    split_location: int,
):
    """Split a cell at the location specified by split_location."""
    manager.split_cell(
        notebook_id=uuid, cell_id=cell_uuid, split_location=split_location
    )
    schedule_notebook_save(background_tasks, uuid=uuid)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{uuid}/undo",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def undo_notebook_operation(
    uuid: UUID, background_tasks: BackgroundTasks
):
    "Undo the last operaton on the notebook with given UUID."
    manager.undo(uuid)
    schedule_notebook_save(background_tasks, uuid=uuid)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{uuid}/redo",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def redo_notebook_operation(
    uuid: UUID, background_tasks: BackgroundTasks
):
    "Redo the last operaton on the notebook with given UUID."
    manager.redo(uuid)
    schedule_notebook_save(background_tasks, uuid=uuid)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/{uuid}/cells/{cell_uuid}",
    status_code=status.HTTP_204_NO_CONTENT,
    dependencies=[Depends(write_access)],
)
async def delete_notebook_cell(
    uuid: UUID, cell_uuid: UUID, background_tasks: BackgroundTasks
):
    manager.delete_cell(uuid, cell_uuid)
    schedule_notebook_save(background_tasks, uuid=uuid)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get("/{uuid}/cells/{cell_uuid}", response_model=CellWrapper)
async def get_notebook_cell(uuid: UUID, cell_uuid: UUID) -> Any:
    cell = manager.find_cell(uuid, cell_uuid)
    return CellWrapper(cell=cell)


@router.get(
    "/{uuid}/cells/{cell_uuid}/execute",
    dependencies=[Depends(write_access)],
    response_model=KernelMessageWrapper,
)
async def execute_notebook_cell(uuid: UUID, cell_uuid: UUID):
    msg_id = await manager.execute(uuid, cell_uuid)
    return KernelMessageWrapper(
        kernel_message=KernelMessage(message_id=msg_id)
    )


@router.get(
    "/{uuid}/interrupt_kernel",
    dependencies=[Depends(write_access)],
    status_code=status.HTTP_204_NO_CONTENT,
)
async def interrupt_kernel(uuid: UUID):
    await manager.interrupt_kernel(uuid)


@router.get(
    "/{uuid}/restart_kernel",
    dependencies=[Depends(write_access)],
    status_code=status.HTTP_204_NO_CONTENT,
)
async def restart_kernel(uuid: UUID):
    await manager.restart_kernel(uuid)


@router.get(
    "/{uuid}/restart_kernel_and_clear_output",
    dependencies=[Depends(write_access)],
    status_code=status.HTTP_204_NO_CONTENT,
)
async def restart_kernel_and_clear_output(uuid: UUID):
    await manager.restart_kernel(uuid, clear_output=True)


@router.get(
    "/{uuid}/restart_kernel_and_execute_all",
    dependencies=[Depends(write_access)],
    response_model=KernelMessagesWrapper,
)
async def restart_kernel_and_execute_all(uuid: UUID):
    msg_ids = await manager.restart_kernel_and_run_all_cells(uuid)
    return KernelMessagesWrapper(
        kernel_messages=[
            KernelMessage(message_id=msg_id) for msg_id in msg_ids
        ]
    )


@router.post(
    "/{uuid}/scratch_execute",
    dependencies=[Depends(write_access)],
    response_model=KernelMessageWrapper,
)
async def execute_notebook_scratch(uuid: UUID, scratch_code: ScratchCode):
    msg_id = await manager.execute_scratch(uuid, scratch_code.code)
    return KernelMessageWrapper(
        kernel_message=KernelMessage(message_id=msg_id)
    )


@router.post(
    "/{uuid}/complete",
    dependencies=[Depends(write_access)],
    response_model=KernelMessageWrapper,
)
async def notebook_code_completion(uuid: UUID, code: CodeComplete):
    msg_id = await manager.complete(uuid, code.code, code.cursor_position)
    return KernelMessageWrapper(
        kernel_message=KernelMessage(message_id=msg_id)
    )


@router.post("/{uuid}/history", response_model=KernelMessageWrapper)
async def notebook_history(uuid: UUID):
    msg_id = await manager.history(uuid)
    return KernelMessageWrapper(
        kernel_message=KernelMessage(message_id=msg_id)
    )


@router.get(
    "/{uuid}/vars",
    dependencies=[Depends(write_access)],
    response_model=KernelVariablesWrapper,
)
async def get_notebook_vars(uuid: UUID):
    """Get a list of vars defined in the kernel."""
    vars = await manager.get_vars(uuid)
    return KernelVariablesWrapper(vars=vars)


@router.get(
    "/{uuid}/vars/{var_name}",
    dependencies=[Depends(write_access)],
    response_model=KernelVariableWrapper,
)
async def get_notebook_single_var(uuid: UUID, var_name: str):
    """Get details on a single var defined in the kernel."""
    single_var = await manager.get_kernel_var_details(uuid, var_name)
    if single_var is None:
        raise HTTPException(404, f"Variable {var_name} not found")
    return KernelVariableWrapper(single_var=single_var)


@router.patch(
    "/{uuid}/cells/{cell_uuid}",
    response_model=CellWrapper,
    dependencies=[Depends(write_access)],
)
async def patch_notebook_cell(
    uuid: UUID,
    cell_uuid: UUID,
    cell_update: CellUpdate,
    background_tasks: BackgroundTasks,
) -> Any:
    cell = manager.update_cell(uuid, cell_uuid, cell_update)
    schedule_notebook_save(background_tasks, uuid=uuid)
    return CellWrapper(cell=cell)


@router.patch(
    "/{uuid}/cells/{cell_uuid}/execute",
    dependencies=[Depends(write_access)],
    response_model=CellKernelMessageWrapper,
)
async def patch_and_execute_notebook_cell(
    uuid: UUID,
    cell_uuid: UUID,
    cell_update: CellUpdate,
    background_tasks: BackgroundTasks,
):
    cell = manager.update_cell(uuid, cell_uuid, cell_update)
    msg_id = await manager.execute(uuid, cell_uuid)
    schedule_notebook_save(background_tasks, uuid=uuid)
    return CellKernelMessageWrapper(
        cell=cell, kernel_message=KernelMessage(message_id=msg_id),
    )


@router.patch(
    "/{uuid}/change_working_directory",
    dependencies=[Depends(write_access)],
    response_model=KernelMessageWrapper,
)
async def change_working_directory(uuid: UUID, directory: str):
    if not path.exists(directory):
        raise HTTPException(404, "Directory does not exist")
    msg_id = await manager.change_kernel_workdir(uuid, directory=directory)
    return KernelMessageWrapper(
        kernel_message=KernelMessage(message_id=msg_id)
    )


def format_file_path(
    directory: Optional[str], filename: str, allow_exists: bool = False,
) -> Dict[str, str]:
    """
    Gets absolute path for file, filename with extension, checks that directory
    exists, and makes sure file doesn't already exist.
    """
    filedir = Path(settings.ROOT_DIR)
    file_path_obj = Path(filename).with_suffix(".ipynb")
    relative_path = file_path_obj
    if directory is not None:
        directory_path = Path(directory)
        if directory_path.is_absolute():
            filedir = directory_path
        else:
            filedir = filedir / directory_path
        relative_path = directory_path / relative_path
    if not filedir.exists():
        raise HTTPException(
            status_code=404, detail=f"Directory does not exist {str(filedir)}"
        )
    filepath = filedir / file_path_obj
    if not allow_exists and filepath.exists():
        raise HTTPException(status_code=400, detail=f"File already exists")
    return {
        "directory": str(filedir),
        "filename": file_path_obj.stem,
        "relative_path": str(relative_path),
        "absolute_path": str(filepath),
    }


@router.post(
    "/upload",
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(write_access)],
    response_model=NotebookPath,
)
async def upload_notebook(
    request: Request, filename: str, directory: Optional[str] = None
) -> Any:
    body = await request.body()
    path_info = format_file_path(directory, filename)
    try:
        manager.upload_notebook(
            body, path_info["directory"], path_info["filename"]
        )
    except (jsonschema.exceptions.ValidationError, AttributeError) as e:
        raise HTTPException(400, f"Failed to parse notebook: {type(e)} {e}")
    return NotebookPath(path=path_info["relative_path"])


@router.post(
    "/upload_and_open",
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(write_access)],
    response_model=NotebookWrapper,
)
async def upload_and_open_notebook(
    request: Request,
    filename: Optional[str] = None,
    directory: Optional[str] = None,
    kspec_name: Optional[str] = None,
    working_directory: Optional[str] = None,
    autosave: bool = True,
) -> Any:
    body = await request.body()

    if filename is not None:
        # If not autosaving allow the file to exists
        path_info = format_file_path(
            directory, filename, allow_exists=not autosave
        )
        directory = path_info["directory"]
        filename = path_info["filename"]
    if directory is None:
        directory = settings.ROOT_DIR
    try:
        nb_uuid = await manager.add_notebook_json(
            body,
            kspec_name=kspec_name,
            directory=directory,
            filename=filename,
            working_directory=working_directory,
            autosave=autosave,
        )

        nb = manager.get_notebook_node(nb_uuid)
        return NotebookWrapper(notebook=nb.dict())
    except (jsonschema.exceptions.ValidationError, AttributeError) as e:
        raise HTTPException(400, f"Failed to parse notebook: {type(e)} {e}")


@router.get(
    "/open/",
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(write_access)],
    response_model=NotebookWrapper,
)
async def open_notebook(
    filepath: str,
    kspec_name: Optional[str] = None,
    working_directory: Optional[str] = None,
):
    abs_path = (
        f"{settings.ROOT_DIR}/{filepath}"
        if not filepath.startswith("/")
        else filepath
    )
    if not path.exists(abs_path):
        raise HTTPException(status_code=404, detail=f"File not found")
    try:
        nb_uuid = await manager.open_notebook(
            abs_path, kspec_name=kspec_name, working_directory=working_directory
        )

        nb = manager.get_notebook_node(nb_uuid)
        return NotebookWrapper(notebook=nb.dict())

    except (jsonschema.exceptions.ValidationError, AttributeError) as e:
        raise HTTPException(400, f"Failed to parse notebook: {type(e)} {e}")
    

@router.post(
    "/create",
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(write_access)],
    response_model=NotebookWrapper,
)
async def create_notebook(
    request: Request,
    filename: str,
    directory: Optional[str] = None,
    kspec_name: Optional[str] = None,
    working_directory: Optional[str] = None,
):
    path_info = format_file_path(directory, filename)
    nb = await manager.create_notebook(
        kspec_name=kspec_name,
        directory=path_info["directory"],
        filename=path_info["filename"],
        working_directory=working_directory,
    )
    return NotebookWrapper(notebook=nb.dict())


@router.patch(
    "/{uuid}/save_as",
    status_code=status.HTTP_200_OK,
    dependencies=[Depends(write_access)],
    response_model=NotebookPath,
)
async def save_notebook_as(
    uuid: UUID,
    filename: str,
    directory: Optional[str] = None,
    overwrite: bool = False,
):
    path_info = format_file_path(directory, filename, overwrite)
    manager.save_notebook_as(
        uuid, path_info["directory"], path_info["filename"]
    )
    return NotebookPath(path=path_info["absolute_path"])


@router.patch(
    "/{uuid}/rename",
    status_code=status.HTTP_200_OK,
    dependencies=[Depends(write_access)],
    response_model=NotebookPath,
)
async def rename_notebook(
    uuid: UUID,
    filename: str,
    directory: Optional[str] = None,
    overwrite: bool = False,
):
    path_info = format_file_path(directory, filename, overwrite)
    manager.save_notebook_as(
        uuid, path_info["directory"], path_info["filename"], delete_old=True
    )
    return NotebookPath(path=path_info["absolute_path"])


@router.websocket("/{uuid}/ws/stream")
async def websocket_stream(
    uuid: UUID,
    websocket: WebSocket,
    read_access: bool = Depends(read_access_websocket),
):
    await websocket.accept()
    msgs = []

    async def receive_channel_message(
        sender, msg, kernel_id, channel, **kwargs
    ):
        if kernel_id == uuid:
            # Add the 'channel' attribute since we're multiplex all of this
            # on one websocket (like the jupyter-kernel-gateway)
            msg["channel"] = channel

            # add the cell id to the metadata if avaiable
            parent_id = None
            if msg.get("parent_header") is not None:
                parent_id = msg["parent_header"].get("msg_id")
            if parent_id is not None:
                id_tuple = manager.notebook_cell_id_for_message_id(parent_id)
                if id_tuple is not None:
                    (nb_id, cell_id) = id_tuple
                    add_to_dict_metadata(msg, cell_uuid=cell_id)

            json_string = json.dumps(msg, cls=D1Encoder)
            msgs.append(json_string)

    signal(IOPUB_CHANNEL).connect(receive_channel_message)
    signal(SHELL_CHANNEL).connect(receive_channel_message)
    signal(STDIN_CHANNEL).connect(receive_channel_message)
    signal(HB_CHANNEL).connect(receive_channel_message)
    signal(CONTROL_CHANNEL).connect(receive_channel_message)

    await websocket_poll(websocket, msgs)


@router.websocket("/{uuid}/ws/notebook")
async def websocket_cells(
    uuid: UUID,
    websocket: WebSocket,
    write_access: bool = Depends(write_access_websocket),
):
    await websocket.accept()
    cell_update_queue: List[str] = []

    async def receive_cell_update(sender, cells, notebook_id, **kwargs):
        logger.debug("receiving CELL_UPDATE")
        if notebook_id == uuid:
            wrapper_dict = {"cell_update": cells}
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(CELL_UPDATE).connect(receive_cell_update)

    async def receive_cell_added(sender, cells, notebook_id, **kwargs):
        logger.debug("receiving CELL_ADDED")
        if notebook_id == uuid:
            wrapper_dict = {"cell_add": cells}
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(CELL_ADDED).connect(receive_cell_added)

    async def receive_cell_deleted(sender, cells, notebook_id, **kwargs):
        logger.debug("receiving CELL_DELETED")
        if notebook_id == uuid:
            wrapper_dict = {"cell_delete": cells}
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(CELL_DELETED).connect(receive_cell_deleted)

    async def receive_cell_stream(
        sender, output, notebook_id, cell_id, **kwargs
    ):
        logger.debug("receiving CELL_STREAM")
        if notebook_id == uuid:
            wrapper_dict = {
                "cell_stream": {"output": output, "cell_id": cell_id}
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(CELL_STREAM).connect(receive_cell_stream)

    async def receive_cell_execution_reply(
        sender, execution_count, notebook_id, cell_id, parent_id, **kwargs
    ):
        logger.debug("receiving CELL_EXECUTION_REPLY")
        if notebook_id == uuid:
            wrapper_dict = {
                "cell_execution_reply": {
                    "cell_id": cell_id,
                    "execution_count": execution_count,
                    "parent_id": parent_id,
                }
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(CELL_EXECUTION_REPLY).connect(receive_cell_execution_reply)

    async def receive_cell_execution_input(
        sender, notebook_id, parent_id, content, **kwargs
    ):
        logger.debug("receiving CELL_EXECUTION_INPUT")
        if notebook_id == uuid:
            wrapper_dict = {
                "cell_execution_input": {
                    "content": content,
                    "parent_id": parent_id,
                }
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(CELL_EXECUTION_INPUT).connect(receive_cell_execution_input)

    async def receive_cell_execution_request(
        sender, notebook_id, cell_id, msg_id, **kwargs
    ):
        logger.debug("receiving CELL_EXECUTION_REQUEST")
        if notebook_id == uuid:
            wrapper_dict = {
                "cell_execution_request": {
                    "cell_id": cell_id,
                    "message_id": msg_id,
                }
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(CELL_EXECUTION_REQUEST).connect(receive_cell_execution_request)

    async def receive_scratch_update(
        sender: Any,
        msg_id: UUID,
        notebook_id: UUID,
        output: Optional[NotebookNode] = None,
        execution_state: Optional[str] = None,
        **kwargs,
    ):
        logger.debug("receiving SCRATCH_UPDATE")
        if notebook_id == uuid:
            scratch_update: Dict[str, Any] = {"message_id": msg_id}
            if output is not None:
                scratch_update["output"] = output
            if execution_state is not None:
                scratch_update["execution_state"] = execution_state
            wrapper_dict = {"scratch_update": scratch_update}
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(SCRATCH_UPDATE).connect(receive_scratch_update)

    async def receive_vars_update(
        sender: Any, notebook_id: UUID, vars: List[KernelVariable], **kwargs
    ):
        logger.debug("receiving VARS_UPDATE")
        if notebook_id == uuid:
            wrapper_dict = {"vars": jsonable_encoder(vars)}
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(VARS_UPDATE).connect(receive_vars_update)

    async def receive_page_payload(
        sender: Any,
        msg_id: UUID,
        notebook_id: UUID,
        data: Dict[str, Any],
        start: Optional[int],
        **kwargs,
    ):
        logger.debug("receiving PAYLOAD_PAGE")
        if notebook_id == uuid:
            wrapper_dict = {
                "payload_page": {
                    "data": data,
                    "start": start,
                    "message_id": msg_id,
                }
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(PAYLOAD_PAGE).connect(receive_page_payload)

    async def receive_code_complete_reply(
        sender: Any,
        msg_id: UUID,
        notebook_id: UUID,
        matches: List[str],
        cursor_start: int,
        cursor_end: int,
        **kwargs,
    ):
        logger.debug("receiving COMPLETE_REPLY")
        if notebook_id == uuid:
            wrapper_dict = {
                "complete_reply": {
                    "matches": matches,
                    "cursor_start": cursor_start,
                    "cursor_end": cursor_end,
                    "message_id": msg_id,
                }
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(COMPLETE_REPLY).connect(receive_code_complete_reply)

    async def receive_history_reply(
        sender: Any,
        msg_id: UUID,
        notebook_id: UUID,
        history: List[Any],
        **kwargs,
    ):
        logger.debug("receiving HISTORY_REPLY")
        if notebook_id == uuid:
            wrapper_dict = {
                "history_reply": {
                    "history": [
                        {
                            "line_number": item[1],
                            "input": item[2][0],
                            "output": item[2][1],
                        }
                        for item in history
                    ],
                    "message_id": msg_id,
                }
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(HISTORY_REPLY).connect(receive_history_reply)

    async def receive_metadata_update_reply(
        sender: Any, notebook_id: UUID, metadata: NotebookNode, **kwargs
    ):
        logger.debug("receiving METADATA_UPDATE")
        if notebook_id == uuid:
            wrapper_dict = {"metadata": metadata}
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(METADATA_UPDATE).connect(receive_metadata_update_reply)

    async def receive_kernel_restarted(
        sender: Any,
        cells: List[NotebookNode],
        notebook_id: UUID,
        run_all_cells: bool,
        **kwargs,
    ):
        logger.debug("receiving KERNEL_RESTARTED")
        if notebook_id == uuid:
            wrapper_dict = {
                "kernel_restarted": {
                    "run_all_cells": run_all_cells,
                    "cells": cells,
                }
            }
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(KERNEL_RESTARTED).connect(receive_kernel_restarted)

    async def receive_kernel_interrupted(
        sender: Any, notebook_id: UUID, **kwargs,
    ):
        logger.debug("receiving KERNEL_INTERRUPTED")
        if notebook_id == uuid:
            wrapper_dict = {"kernel_interrupted": {"interrupted": True}}
            wrapper_string = json.dumps(wrapper_dict, cls=D1Encoder)
            cell_update_queue.append(wrapper_string)

    signal(KERNEL_INTERRUPTED).connect(receive_kernel_interrupted)

    await websocket_poll(websocket, cell_update_queue)
